# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np


def optimize(job, cluster, max_cluster_size):
    hints = job.hints
    if not hints:
        return ["adaptdl_virtual_node_0"]

    job_info = job.job_info

    existing_nodes = cluster.get_nodes()
    virtual_nodes = [
        f"adaptdl_virtual_node_{i}"
        for i in range(max_cluster_size - len(existing_nodes))]

    existing_ips = [node["NodeManagerAddress"] for node in existing_nodes]

    nodes = existing_ips + virtual_nodes

    existing_ips = set(existing_ips)
    node_resources = {
        node["NodeManagerAddress"]: node["Resources"]
        for node in existing_nodes}

    job_resources = cluster.worker_resources

    replicas_per_node = [0 for node in nodes]
    for index, node in enumerate(nodes):
        if node not in node_resources:
            resources = cluster.worker_resources
        else:
            resources = node_resources[node]
        replicas_per_node[index] = int(
            min(resources.get(resource_type, 0.0) / value
                for resource_type, value in job_resources.items()))

    max_workers = sum(replicas_per_node)

    allocation = []

    workers_left = max_workers

    while workers_left > 0:
        count = (min(replicas_per_node[0], workers_left))
        if count:
            allocation += [(nodes[0]) * count]
            if count == replicas_per_node[0]:
                nodes = nodes[1:]
                replicas_per_node = replicas_per_node[1:]
            workers_left -= count
        else:
            nodes = nodes[1:]
            replicas_per_node = replicas_per_node[1:]

    speedup_fn = job_info.speedup_fn
    base_speedup = speedup_fn(1, 1)

    workers_arr = np.asarray(range(1, max_workers+1))
    speedups = speedup_fn(workers_arr, workers_arr)

    best_replicas = 0
    best_speedup = 0.0
    nodes_used = set()
    current_replicas = len(job._worker_tasks)

    current_speedup = speedups[current_replicas - 1]
    for worker, speedup in enumerate(speedups):
        nodes_used.add(allocation[worker])
        num_nodes = len(nodes_used)
        if (speedup / num_nodes >= base_speedup * 0.5):
            best_replicas = worker + 1
            best_speedup = speedup

    if (best_speedup < current_speedup * 1.05 or
            (abs(best_replicas + 1 - current_replicas) <
             0.15 * current_replicas)):
        best_replicas = current_replicas

    return allocation[:best_replicas]



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import importlib
import logging
import os
from pathlib import Path
import pkg_resources
import requests
import sys
import time
import traceback

from adaptdl_ray.aws.utils import \
    _checkpoint_obj_to_dir, _serialize_checkpoint, Status

import ray
import ray._private.services as services


@ray.remote(num_cpus=0.1, max_retries=0)
def listen_for_spot_termination(timeout=None):
    MOCK = (os.environ.get("MOCK", "False").lower() == "true")
    logging.basicConfig(level=logging.INFO)

    if MOCK:
        logging.debug("Using mocked spot instance")
        endpoint = f"{services.get_node_ip_address()}:8234"
    else:
        # AWS spot instance termination endpoint
        endpoint = "169.254.169.254"

    start = time.time()

    while True:
        try:
            resp = requests.get(
                f'http://{endpoint}/latest/meta-data/spot/instance-action',
                timeout=0.1)
            if resp.status_code == 404:
                # AWS endpoint responded, no termination detected
                time.sleep(5)
            elif resp.status_code >= 200 and resp.status_code < 300:
                resp_json = resp.json()
                if (resp_json["action"] == "terminate"
                        or resp_json["action"] == "stop"):
                    ip = services.get_node_ip_address()
                    logging.info(f"termination detected on node {ip}")
                    return ip
            else:
                raise RuntimeError(
                    "AWS spot instance interrupt warning "
                    "endpoint not responding")
            if timeout and time.time() - start > timeout:
                return None
        except requests.RequestException as e:
            logging.error(e)
            time.sleep(5)


@ray.remote(max_retries=0)
def run_adaptdl(job_key, job_uid, rank, replicas,
                num_restarts, checkpoint=None, offset=0, path="", argv=None):
    logging.basicConfig(level=logging.INFO)
    logging.info(f"Starting worker {rank}")

    def report_status(status):
        status_obj_ref = ray.put(status.value)
        controller.register_status.remote(status_obj_ref)

    controller = ray.get_actor("AdaptDLController")
    supervisor_url = ray.get(controller.get_url.remote())

    os.environ["ADAPTDL_MASTER_PORT"] = str(47000 + num_restarts + offset)
    os.environ["ADAPTDL_REPLICA_RANK"] = str(rank)
    os.environ["ADAPTDL_NUM_REPLICAS"] = str(replicas)
    os.environ["ADAPTDL_SUPERVISOR_URL"] = supervisor_url
    os.environ["ADAPTDL_JOB_ID"] = job_key
    os.environ["ADAPTDL_NUM_RESTARTS"] = str(num_restarts)
    os.environ["ADAPTDL_SCHED_VERSION"] = str(
        pkg_resources.get_distribution("adaptdl").version)
    suffix = f"{job_uid}-{rank}"
    checkpoint_path = f"/tmp/checkpoint-{suffix}"

    try:
        if os.path.exists(checkpoint_path):
            import shutil
            shutil.rmtree(checkpoint_path)
        os.mkdir(checkpoint_path)
        if checkpoint:
            _checkpoint_obj_to_dir(checkpoint_path, checkpoint)
        num_restarts = int(num_restarts)
        os.environ["ADAPTDL_CHECKPOINT_PATH"] = str(checkpoint_path)
        share_path = f"/tmp/share-{suffix}"
        if not os.path.exists(share_path):
            os.mkdir(share_path)
        os.environ["ADAPTDL_SHARE_PATH"] = str(share_path)

        rank_obj_ref = ray.put(rank)
        ip_obj_ref = ray.put(services.get_node_ip_address())
        controller.register_worker.remote(rank_obj_ref, ip_obj_ref)
    except Exception as e:
        logging.info(traceback.format_exc())
        time.sleep(5)
        report_status(Status.FAILED)
        raise e

        # TODO: replace with block
    try:
        filename = Path(path).name
        sys.argv = [filename]
        if argv:
            # Need to augment the argv to mimic that file being called
            sys.argv += argv
        spec = importlib.util.spec_from_file_location("__main__", path)
        module = importlib.util.module_from_spec(spec)
        # TODO: fix imports when caller module is not in the root path
        spec.loader.exec_module(module)
        time.sleep(5)

    except SystemExit:
        # Received a cancel from the controller -- the job is being rescheduled
        # Worker 0 needs to send the checkpoint back to the controller so the
        # next generation of workers can resume
        logging.info(f"Worker {rank} received system exit")
        if rank == 0:
            checkpoint_obj = _serialize_checkpoint(checkpoint_path)
            logging.info("checkpoint created")
            checkpoint_obj_ref = ray.put(checkpoint_obj)
            logging.info("checkpoint placed")
            result = ray.get(
                controller.register_checkpoint.remote(checkpoint_obj_ref))
            logging.info(f"checkpoint registered: {result}")
        # This sleep is to keep this remote task alive
        # until its worker object can be killed by the controller
        time.sleep(1800)

    except Exception as e:
        logging.error(traceback.format_exc())
        logging.error(e)
        time.sleep(5)
        report_status(Status.FAILED)
        raise e
    else:
        if rank == 0:
            logging.info("Job succeeded, exiting")
            time.sleep(5)
            report_status(Status.SUCCEEDED)
            time.sleep(5)



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import torch.distributed as dist
import ray

from .adaptdl_trainable import AdaptDLTrainableCreator, _train_simple


@pytest.fixture
def ray_start_2_cpus():
    address_info = ray.init(num_cpus=2)
    yield address_info
    # The code after the yield will run as teardown code.
    ray.shutdown()
    # Ensure that tests don't ALL fail
    if dist.is_initialized():
        dist.destroy_process_group()


def test_single_step(ray_start_2_cpus):  # noqa: F811
    trainable_cls = AdaptDLTrainableCreator(_train_simple, num_workers=2)
    trainer = trainable_cls()
    trainer.train()
    trainer.stop()


def test_step_after_completion(ray_start_2_cpus):  # noqa: F811
    trainable_cls = AdaptDLTrainableCreator(_train_simple, num_workers=2)
    trainer = trainable_cls(config={"epochs": 1})
    with pytest.raises(RuntimeError):
        for i in range(10):
            trainer.train()


def test_save_checkpoint_restore(ray_start_2_cpus):  # noqa: F811
    trainable_cls = AdaptDLTrainableCreator(_train_simple, num_workers=2)
    trainer = trainable_cls(config={"epochs": 1})
    trainer.train()
    checkpoint_dir = trainer.save()
    trainer.restore(checkpoint_dir)
    trainer.train()
    trainer.stop()



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from itertools import cycle
from typing import Dict, List
from adaptdl_sched.policy.pollux import PolluxPolicy
from adaptdl_sched.policy.utils import NodeInfo
from adaptdl_ray.adaptdl.adaptdl_job_mixin import AdaptDLJobMixin
from adaptdl_ray.adaptdl import config


class AdaptDLAllocator:
    def __init__(self, nodes: List = None):
        nodes = nodes if nodes is not None else config.nodes()
        self._node_infos = {node['NodeManagerAddress']:
                            NodeInfo(node['Resources'], preemptible=False)
                            for node in nodes}
        self._default_node = cycle(list(self._node_infos))
        # Add a node template.
        self._node_template = NodeInfo(list(self._node_infos.values())[0].
                                       resources, preemptible=False)
        self._policy = PolluxPolicy()

    def default_allocation(self, num_devices=1) -> List[str]:
        """ Cycle through nodes for default trial allocation."""
        return [f"{next(self._default_node)}"] * num_devices

    def allocate(self,
                 jobs: List[AdaptDLJobMixin],
                 nodes: List = None) -> (Dict, int):
        """ Use Pollux to distribute available resources between jobs."""
        if nodes is not None:
            node_infos = {node['NodeManagerAddress']:
                          NodeInfo(node['Resources'], preemptible=False)
                          for node in nodes}
        else:
            node_infos = self._node_infos

        assert len(jobs) > 0
        # gather JobInfos
        job_infos = {job.job_id: job.job_info for job in jobs}
        # gather previous allocations
        prev_allocs = {job.job_id: job.allocation for job in jobs}

        allocations, desired_nodes = \
            self._policy.optimize(job_infos,
                                  node_infos,
                                  prev_allocs,
                                  self._node_template)
        # Fill empty allocations for jobs which didn't get any
        for job_id in job_infos:
            allocations[job_id] = allocations.get(job_id, [])

        assert all(v == [] for k, v in allocations.items()) is False
        return allocations, desired_nodes



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import ray
import pytest
from adaptdl_ray.aws.worker import run_adaptdl, listen_for_spot_termination
import asyncio
import os
import time

from aiohttp import web


# see https://stackoverflow.com/a/60334747
@pytest.fixture(scope="module")
def ray_fix():
    if not ray.is_initialized:
        ray.init()
        yield None
        ray.shutdown()
    else:
        yield None


@ray.remote
class MockedController():

    def __init__(self):
        self.checkpoint = None

    def register_status(self, *args, **kwargs):
        print(*args, **kwargs)
        pass

    def register_worker(self, *args, **kwargs):
        print(*args, **kwargs)
        pass

    def register_checkpoint(self, obj):
        self.checkpoint = obj

    def get_checkpoint(self):
        return self.checkpoint

    def get_url(self):
        return ray._private.services.get_node_ip_address()


@ray.remote
class TerminationEndpoint():
    def __init__(self):
        self.terminating = False

    async def handle_termination_request(self, request):
        if self.terminating:
            return web.json_response({"action": "terminate"})
        else:
            return web.HTTPNotFound()

    async def start_server(self):
        app = web.Application()
        app.add_routes([
            web.get('/latest/meta-data/spot/instance-action',
                    self.handle_termination_request)
            ])
        runner = web.AppRunner(app)
        await runner.setup()
        site = web.TCPSite(
            runner, ray._private.services.get_node_ip_address(), 8234)
        await site.start()
        await asyncio.sleep(30)
        await runner.cleanup()

    async def set_to_terminate(self):
        self.terminating = True
        return None


def test_worker(ray_fix):
    controller = MockedController.options(name="AdaptDLController").remote()
    rank = 0
    replicas = 2
    restarts = 3
    checkpoint = None
    offset = 50
    path = "ray/adaptdl_ray/aws/_example_worker.py"
    argv = ["--arg1", "value", "--arg2", "value"]

    worker_task = run_adaptdl.remote(
        "test_key", "test_uid", rank, replicas,
        restarts, checkpoint, offset, path, argv)

    # can't cancel with force=True
    time.sleep(10)
    ray.cancel(worker_task, force=False)
    print("canceling")
    time.sleep(10)
    checkpoint = ray.get(controller.get_checkpoint.remote())
    print(checkpoint)
    assert('file.txt' in checkpoint)
    ray.cancel(worker_task, force=False)

    rank = 1
    replicas = 2
    restarts = 4
    offset = 50

    worker_task = run_adaptdl.remote(
        "test_key_2", "test_uid_2", rank, replicas,
        restarts, checkpoint, offset, path, argv)

    time.sleep(10)
    assert(os.path.exists("/tmp/checkpoint-test_uid_2-1/file.txt"))
    with open("/tmp/checkpoint-test_uid_2-1/file.txt", "rb") as f:
        result = int(f.read())
        assert (result == 5)


async def test_spot_instance_termination(ray_fix):
    endpoint = TerminationEndpoint.remote()
    endpoint.start_server.remote()

    task = listen_for_spot_termination.remote(timeout=5.0)
    ip = ray.get(task)
    assert(ip is None)
    task = listen_for_spot_termination.remote(timeout=15.0)
    time.sleep(5)
    await endpoint.set_to_terminate.remote()
    ip = ray.get(task, timeout=10)
    assert(ip == ray._private.services.get_node_ip_address()), \
        f"found {ip}, expected {ray._private.services.get_node_ip_address()}"



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, Optional
import logging

from ray.tune import trial_runner
from ray.tune.schedulers import TrialScheduler
from ray.tune.trial import Trial

from adaptdl_ray.tune.adaptdl_trial import AdaptDLTrial
from adaptdl_ray.adaptdl import AdaptDLAllocator
from adaptdl_ray.adaptdl import config
from adaptdl_ray.adaptdl.utils import pgs_to_resources

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class AdaptDLScheduler(TrialScheduler):
    """TrialScheduler for the elastic AdaptDL Trials"""
    # Allocator is invoked every _ALLOCATOR_INVOKE_FREQ epochs
    _ALLOCATOR_INVOKE_FREQ = 100

    def __init__(self,
                 allocator: Optional[AdaptDLAllocator] = None,
                 allocator_invoke_freq: Optional[int] = None):
        # Reserve 1 CPU from the first node for the Trainables
        consumed_resources = {config.nodes()[0]["NodeManagerAddress"]:
                              {"CPU": -1.0}}
        self._allocator = allocator if allocator is not None \
            else AdaptDLAllocator(config.nodes(consumed_resources))
        if allocator_invoke_freq is not None:
            self._allocator_invoke_freq = allocator_invoke_freq
        else:
            self._allocator_invoke_freq = AdaptDLScheduler.\
                    _ALLOCATOR_INVOKE_FREQ
        self._allocs = {}
        self._iterations = 0

    @property
    def _should_realloc(self) -> bool:
        """ True if its okay to invoke the allocator."""
        self._iterations += 1
        return (self._iterations %
                self._allocator_invoke_freq == 0)

    def on_trial_add(self, trial_runner: "trial_runner.TrialRunner",
                     trial: Trial):
        """Called after SearchAlgorithm.next_trial"""
        trial = AdaptDLTrial.create_from(trial, trial_runner,
                                         self._allocator.default_allocation())

    def on_trial_error(self, trial_runner: "trial_runner.TrialRunner",
                       trial: Trial):
        pass

    def on_trial_result(self, trial_runner: "trial_runner.TrialRunner",
                        trial: Trial, result: Dict) -> str:
        trials = [trial for trial in trial_runner.get_trials()
                  if trial.status in (Trial.RUNNING, Trial.PENDING)]
        if self._should_realloc and len(self._allocs) == 0:
            in_use_pgs = [pg.to_dict() for pg in
                          trial_runner.trial_executor._pg_manager._in_use_pgs]
            consumed_resources = pgs_to_resources(in_use_pgs)
            nodes = config.nodes(consumed_resources)
            self._allocs, _ = self._allocator.allocate(trials, nodes)

        alloc = self._allocs.pop(trial.trial_id, None)
        if alloc is None:
            # No change in allocation for this Trial
            return TrialScheduler.CONTINUE

        if alloc == [] and trial.status == Trial.RUNNING:
            # Pause only if the trial is running
            trial.pause(trial_runner)
            return TrialScheduler.PAUSE
        elif alloc != trial.allocation:
            trial = AdaptDLTrial.create_from(trial,
                                             trial_runner,
                                             alloc,
                                             copy_state=True)
            # Stop the old trial that's being replaced
            return TrialScheduler.STOP
        return TrialScheduler.CONTINUE

    def on_trial_complete(self, trial_runner: "trial_runner.TrialRunner",
                          trial: Trial, result: Dict):
        pass

    def on_trial_remove(self, trial_runner: "trial_runner.TrialRunner",
                        trial: Trial):
        pass

    def choose_trial_to_run(
            self, trial_runner: "trial_runner.TrialRunner") -> Optional[Trial]:
        for trial in trial_runner.get_trials():
            if (trial.status == Trial.PENDING
                    and trial_runner.trial_executor.
                    has_resources_for_trial(trial)):
                return trial
        for trial in trial_runner.get_trials():
            if (trial.status == Trial.PAUSED and
                    trial_runner.trial_executor.
                    has_resources_for_trial(trial) and
                    len(self._allocs) == 0):
                # Note: this puts the trial back to RUNNING, we allow Trials to
                # resume when the allocation cache is empty and we reach a sync
                # point.
                return AdaptDLTrial.create_from(trial,
                                                trial_runner,
                                                self._allocator.
                                                default_allocation(),
                                                copy_state=True)
        return None

    def debug_string(self) -> str:
        return "Using AdaptDL scheduling algorithm."



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import ray
from ray.cluster_utils import Cluster
from .config import nodes, default_device
from ray.util.placement_group import placement_group


class TrialRunnerTest(unittest.TestCase):
    def setUp(self):
        self.cluster = Cluster(
            initialize_head=True,
            connect=True,
            head_node_args={
                "num_cpus": 4,
                "num_gpus": 1,
            })
        self.cluster.add_node(num_cpus=2, num_gpus=1)
        self.cluster.wait_for_nodes()

    def tearDown(self):
        ray.shutdown()
        self.cluster.shutdown()

    def testAvailableResources(self):
        assert len(nodes()) == 2
        assert default_device(refresh=True) == "GPU"

    def testOthersTakingResources(self):
        # Let someone occupy the head node
        pg = placement_group([{"CPU": 4, "GPU": 1}])
        ray.get(pg.ready())
        # We are left with the second node
        assert len(nodes()) == 1
        assert default_device(refresh=True) == "GPU"

        pg = placement_group([{"GPU": 1}])
        ray.get(pg.ready())
        # Default device should be CPU
        assert default_device(refresh=True) == "CPU"
        assert len(nodes()) == 1



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import unittest

import ray

from ray import tune
from ray.tune.ray_trial_executor import RayTrialExecutor
from ray.tune.suggest import BasicVariantGenerator
from .adaptdl_trainable import AdaptDLTrainableCreator, _train_simple


class TrialRunnerTest(unittest.TestCase):
    def setUp(self):
        # Wait up to five seconds for placement groups when starting a trial
        os.environ["TUNE_PLACEMENT_GROUP_WAIT_S"] = "5"
        # Block for results even when placement groups are pending
        os.environ["TUNE_TRIAL_STARTUP_GRACE_PERIOD"] = "0"
        os.environ["TUNE_TRIAL_RESULT_WAIT_TIME_S"] = "99999"

    def tearDown(self):
        ray.shutdown()

    def testExperimentTagTruncation(self):
        ray.init(num_cpus=2)
        trainable_cls = AdaptDLTrainableCreator(_train_simple, num_workers=1)
        trial_executor = RayTrialExecutor()
        experiments = {
            "foo": {
                "run": trainable_cls.__name__,
                "config": {
                    "a" * 50: tune.sample_from(lambda spec: 5.0 / 7),
                    "b" * 50: tune.sample_from(lambda spec: "long" * 40)
                },
            }
        }

        for name, spec in experiments.items():
            trial_generator = BasicVariantGenerator()
            trial_generator.add_configurations({name: spec})
            while not trial_generator.is_finished():
                trial = trial_generator.next_trial()
                if not trial:
                    break
                trial_executor.start_trial(trial)
                assert len(os.path.basename(trial.logdir)) <= 200
                trial_executor.stop_trial(trial)



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import types
import shutil

from ray.tune.function_runner import wrap_function
from ray.tune.trainable import TrainableUtil

from adaptdl.torch import init_process_group
from adaptdl.checkpoint import save_all_states
from adaptdl.torch._metrics import _get_sched_hints


# Wrap the free functions of AdaptDL with FunctionRunner methods
def save_all_states_remote(self, trial_state):
    """ Save all of AdaptDL's job state and return it as an in-memory
    object."""
    checkpoint = save_all_states()
    parent_dir = TrainableUtil.find_checkpoint_dir(checkpoint)
    checkpoint_path = TrainableUtil.process_checkpoint(checkpoint,
                                                       parent_dir,
                                                       trial_state)
    checkpoint_obj = TrainableUtil.checkpoint_to_object(checkpoint_path)
    # Done with the directory, remove
    shutil.rmtree(checkpoint_path)
    return checkpoint_obj


def get_sched_hints_remote(self):
    """ Return hints for AdaptDL Scheduler."""
    return _get_sched_hints()


def wrap_function_patched(function):
    """ Monkey-patch FunctionRunner remote trainable"""
    func_trainable = wrap_function(function)
    func_trainable.save_all_states = types.MethodType(save_all_states_remote,
                                                      func_trainable)
    func_trainable.get_sched_hints = types.MethodType(get_sched_hints_remote,
                                                      func_trainable)
    return func_trainable


def setup_process_group(*args, **kwargs):
    init_process_group(backend=kwargs["backend"],
                       init_method=kwargs["url"],
                       world_size=kwargs["world_size"],
                       rank=kwargs["world_rank"])



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import time
import adaptdl._signal as signal


def foo(x, y):
    return x + y


if __name__ == "__main__":
    x = foo(3, 2)
    if os.environ.get("ADAPTDL_REPLICA_RANK", None) == "0":
        with open(
                f"{os.environ['ADAPTDL_CHECKPOINT_PATH']}/file.txt", "w") as f:
            f.write(f"{x}\n")
        for _ in range(30):
            time.sleep(1)
            if signal.get_exit_flag():
                exit(143)

else:
    raise RuntimeError("Not running as main")



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List
from datetime import datetime

from adaptdl.goodput import GoodputFunction, GradParams
from adaptdl_sched.policy.speedup import SpeedupFunction
from adaptdl_sched.policy.utils import JobInfo
from adaptdl_ray.adaptdl import config
from adaptdl_ray.adaptdl.utils import pgf_to_allocation


class AdaptDLJobMixin:
    def __init__(self, *args, **kwargs):
        # Be wary of putting large data members here. Tune Experiment
        # checkpointing may try to serialize this.
        self._job_id = kwargs.pop("job_id", 0)
        self.creation_timestamp = kwargs.pop("creation_timestamp",
                                             datetime.now())
        super().__init__(*args, **kwargs)

    @property
    def job_id(self):
        """ Unique job ID assigned to this AdaptDL job"""
        return self._job_id

    def _fetch_metrics(self):
        """ Returns perf metrics of this AdaptDLJob. This could return a cached
        copy in case the job is currently not running."""
        raise NotImplementedError

    def _allocation_in_use(self) -> bool:
        """ Returns True if the allocation is being used by an AdaptDLJob."""
        raise NotImplementedError

    @property
    def job_info(self) -> JobInfo:
        metrics = self._fetch_metrics()
        if metrics is not None:
            perf_params = metrics.perf_params
            if metrics.grad_params is not None:
                grad_params = metrics.grad_params
            else:
                grad_params = GradParams(0.0, 1.0)
            goodput_fn = GoodputFunction(perf_params,
                                         grad_params,
                                         metrics.init_batch_size)
            speedup_fn = SpeedupFunction(goodput_fn,
                                         metrics.max_batch_size,
                                         metrics.local_bsz_bounds,
                                         metrics.gradient_accumulation)
        else:
            speedup_fn = lambda n, r: r  # noqa: E731

        return JobInfo(config.job_resources(),
                       speedup_fn,
                       self.creation_timestamp,
                       config._JOB_MIN_REPLICAS,
                       config._JOB_MAX_REPLICAS)

    @property
    def allocation(self) -> List[str]:
        """ Current allocation the job is utilizing"""
        # Allocation is in use if the job is using it
        if self._allocation_in_use():
            assert self.placement_group_factory is not None
            return pgf_to_allocation(self.placement_group_factory)
        else:
            return []



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import Enum
import os


# Adapted from Ray Tune
def _checkpoint_obj_to_dir(checkpoint_dir, checkpoint_obj):
    for (path, data) in checkpoint_obj.items():
        file_path = os.path.join(checkpoint_dir, path)
        directory = os.path.dirname(file_path)
        os.makedirs(directory, exist_ok=True)
        with open(file_path, "wb") as f:
            f.write(data)
    return


def _serialize_checkpoint(checkpoint_dir):
    data = {}
    for basedir, _, file_names in os.walk(checkpoint_dir):
        for file_name in file_names:
            path = os.path.join(basedir, file_name)
            with open(path, "rb") as f:
                data[os.path.relpath(path, checkpoint_dir)] = f.read()
    return data


class Status(Enum):
    FAILED = 0
    SUCCEEDED = 1
    RUNNING = 2



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from aiohttp import web

import asyncio
import copy
import logging
import uuid
import time

from collections import namedtuple

from adaptdl_ray.adaptdl.adaptdl_job_mixin import AdaptDLJobMixin
from adaptdl_ray.adaptdl.utils import allocation_to_pgf

from adaptdl.goodput import PerfParams, GradParams

import ray
import ray.autoscaler.sdk as sdk
import ray._private.services as services

from .optimizer import optimize
from .utils import Status
from .worker import listen_for_spot_termination, run_adaptdl

from ray.util.placement_group import (
    remove_placement_group
)

# To be used when there are no running tasks
FULL_RESCALE_TIMEOUT = 6000

NAMESPACE = "adaptdl_job"
NAME = "adaptdl_job"

job_uid = str(uuid.uuid4())[:8]


class RayAdaptDLJob(AdaptDLJobMixin):
    def __init__(self,
                 worker_resources, worker_port_offset,
                 checkpoint_timeout, **kwargs):
        super().__init__(job_id=job_uid)
        self._worker_resources = worker_resources
        self._job_params = kwargs

        # TODO: merge workers, worker_tasks
        self._workers = {}  # int: IP
        self._worker_tasks = {}
        self._last_metrics = None

        self._runner = None
        self._pg = None
        self._worker_port_offset = worker_port_offset

        self._running = False
        self._checkpoint = None
        self._checkpoint_ref = None
        self._checkpoint_received = asyncio.Event()
        self._checkpoint_lock = asyncio.Lock()
        self._checkpoint_timeout = checkpoint_timeout

        self._iteration = 0

        self.completed = asyncio.Event()
        self._status = Status.RUNNING

        self._placement_group_factory = None

    async def force_worker_checkpoint(self):
        if self._worker_tasks:
            logging.info("Checkpoint needed: stopping workers")
        async with self._checkpoint_lock:
            if self._running:
                for index, task in self._worker_tasks.items():
                    try:
                        ray.cancel(task, force=False)
                    except Exception as e:
                        raise e
                self._running = False

    def _stop_workers(self):
        if self._worker_tasks:
            logging.info("Terminating workers")
        for index, task in self._worker_tasks.items():
            try:
                ray.cancel(task, force=True)
            except Exception:
                pass
        self._worker_tasks = {}
        self._workers = {}

    @property
    def placement_group_factory(self):
        return self._placement_group_factory

    def register_hints(self, hints):
        self._last_metrics = copy.deepcopy(hints)

    @property
    def status(self):
        return self._status

    @property
    def worker_resources(self):
        return copy.deepcopy(self._worker_resources)

    @property
    def workers(self):
        return copy.deepcopy(self._workers)

    @property
    def hints(self):
        hints = copy.deepcopy(self._last_metrics)
        if hints:
            if hints["gradParams"]:
                hints["gradParams"]["sqr"] = hints["gradParams"]["norm"]
                del hints["gradParams"]["norm"]
            else:
                hints["gradParams"] = {"sqr": 1.0, "var": 1.0}
            hints["gradParams"] = GradParams(**hints["gradParams"])
            hints["perfParams"] = PerfParams(**hints["perfParams"])
        return hints

    def _fetch_metrics(self):
        hints = self.hints
        metrics = {}
        for key, value in hints.items():
            metrics_key = ''.join(
                ['_'+c.lower() if c.isupper() else c for c in key]).lstrip('_')
            metrics[metrics_key] = value
        Metrics = namedtuple("Metrics", metrics.keys())
        return Metrics(**metrics)

    def _allocation_in_use(self):
        return self._workers

    async def _create_workers(self, allocation):
        if self._pg:
            remove_placement_group(self._pg)
        logging.info(f"Creating {len(allocation)} worker tasks")
        self._placement_group_factory = allocation_to_pgf(
            allocation, self._worker_resources)
        self._pg = self._placement_group_factory()
        self._worker_tasks = {
            worker_index:
            run_adaptdl.options(
                num_cpus=self._worker_resources.get("CPU", 1),
                num_gpus=self._worker_resources.get("GPU", 0),
                placement_group=self._pg).remote(
                    f"{NAMESPACE}/{NAME}",
                    job_uid,
                    worker_index,
                    len(allocation),
                    self._iteration,
                    self._checkpoint_ref,
                    self._worker_port_offset,
                    **self._job_params)
            for worker_index, node in enumerate(allocation)}

        self._checkpoint_received.clear()
        self._running = True
        self._iteration += 1

    async def update_workers(self, allocation):
        logging.info(
            f"Updating workers. before: {len(self._workers.values())} workers"
            f", after: {len(allocation)} workers")
        if set(self._workers.values()) != set(allocation):
            await self.force_worker_checkpoint()
            # TODO: use asyncio events
            if self._worker_tasks:
                try:
                    await asyncio.wait_for(
                        self._checkpoint_received.wait(),
                        self._checkpoint_timeout)
                except asyncio.TimeoutError:
                    logging.warning("Waited for checkpoint, not found. "
                                    "Proceeding with previous checkpoint")
            self._stop_workers()
            await self._create_workers(allocation)
            return list(self._worker_tasks.values())
        else:
            logging.info("Allocation unchanged, proceeding")
            return None

    def register_checkpoint(self, obj):
        self._checkpoint = obj
        self._checkpoint_received.set()
        self._checkpoint_ref = ray.put(obj)

    def register_status(self, status):
        if self._status != Status.SUCCEEDED:
            self._status = Status(status)
        if self._status != Status.RUNNING:
            self.completed.set()


class Cluster():
    def __init__(self, worker_resources, rescale_timeout):
        self._worker_resources = worker_resources
        self._rescale_timeout = rescale_timeout
        self._terminating_nodes = set()
        self._force_immediate_allocation = asyncio.Event()

    @property
    def worker_resources(self):
        return copy.deepcopy(self._worker_resources)

    def mark_node_for_termination(self, node):
        self._terminating_nodes.add(node)

    @property
    def _invalid_nodes(self):
        current_ip = services.get_node_ip_address()
        return self._terminating_nodes.union({current_ip})

    def get_nodes(self):
        nodes = ray.nodes()
        return [node for node in nodes
                if (node["NodeManagerAddress"] not in self._invalid_nodes
                    and node["alive"] and "Resources" in node)]

    def _cluster_ready(self, allocation):
        nodes = self.get_nodes()
        nodes = {
            node["NodeManagerAddress"]: node["Resources"] for node in nodes}

        virtual_workers = []

        found_workers_count = 0
        for worker in allocation:
            if ("adaptdl_virtual" in worker or worker not in nodes):
                virtual_workers += [worker]
            else:
                node = nodes[worker]
                for resource, amount in self._worker_resources.items():
                    node[resource] -= amount
                found_workers_count += 1

        for _ in virtual_workers:
            for node in nodes.values():
                if all([node.get(resource, 0.0) >= amount
                        for resource, amount
                        in self._worker_resources.items()]):
                    for resource, amount in self._worker_resources.items():
                        if node[resource]:
                            node[resource] -= amount
                    found_workers_count += 1
                    break
            else:
                return False, found_workers_count
        return found_workers_count >= len(allocation), found_workers_count

    async def expand_cluster(self, current_workers, allocation):
        logging.info("Attempting to expand cluster to "
                     f"{len(allocation)} nodes")
        nodes = self.get_nodes()
        node_ips = {node["NodeManagerAddress"] for node in nodes}
        invalid_workers = {
            worker_id: ip for worker_id, ip in current_workers.items()
            if ip not in node_ips}
        if len(invalid_workers) == len(current_workers):
            rescale_timeout = FULL_RESCALE_TIMEOUT
            logging.info(
                "No live workers found. "
                "Waiting longer than specified for rescaling.")
        else:
            rescale_timeout = self._rescale_timeout

        worker_resources = [
            copy.deepcopy(self._worker_resources)
            for _ in range(len(allocation) + len(invalid_workers))]
        for bundle in worker_resources:
            bundle["CPU"] += 0.1
        sdk.request_resources(bundles=worker_resources)
        waited = 0.0
        logging.info(f"Waiting for up to {rescale_timeout} seconds for "
                     "nodes to be ready")
        while (waited < rescale_timeout and
               not self._cluster_ready(allocation)[0] and
               not self._force_immediate_allocation.is_set()):
            await asyncio.sleep(1.0)
            waited += 1.0
        ready, nodes = self._cluster_ready(allocation)
        logging.info(f"Found {nodes} available nodes")
        if not ready:
            allocation = (
                [node for node in allocation
                 if "adaptdl_virtual" not in node] +
                [node for node in allocation if "adaptdl_virtual" in node])
            return allocation[:nodes]
        else:
            return allocation


# This class is generally used remotely, but for testing purposes
# we also keep a local version. See https://stackoverflow.com/a/62309671
# from Robert Nishihara
class Controller():
    def __init__(self, cluster_size, rescale_timeout):
        super().__init__()
        logging.basicConfig(level=logging.INFO)
        self._job = None
        self._cluster = None
        self._cluster_size = cluster_size
        self._runner = None
        self._url = f"http://{services.get_node_ip_address()}:8080"
        self._spot_listener_tasks = {}
        self._ready = asyncio.Event()
        self._completed = asyncio.Event()
        self._rescale_timeout = rescale_timeout
        self._reschedule_queue = asyncio.Queue(maxsize=1)
        self._last_deployed = time.time()

    def get_url(self):
        return self._url

    async def run_controller(self):
        asyncio.create_task(self._run_app())
        asyncio.create_task(self._reschedule_listener())
        await self._completed.wait()
        await self._runner.cleanup()

    async def create_job(self, *args, **kwargs):
        await self._ready.wait()
        self._job = RayAdaptDLJob(*args, **kwargs)
        self._cluster = Cluster(
            self._job.worker_resources,
            self._rescale_timeout)
        await self._enqueue_reschedule(immediate=True)
        await self._job.completed.wait()
        self._completed.set()
        return self._job.status

    async def _reschedule_listener(self):
        while True:
            immediate = await self._reschedule_queue.get()
            logging.info("waiting for job")
            current_time = time.time()
            if not immediate and (current_time - self._last_deployed <= 300):
                await asyncio.sleep(int(current_time - self._last_deployed))
            await self._reschedule_jobs()
            self._last_deployed = time.time()
            logging.info("done rescheduling")

    async def _enqueue_reschedule(self, immediate=False):
        if not immediate:
            try:
                self._reschedule_queue.put_nowait(False)
                logging.info("enqueued reschedule")
            except asyncio.QueueFull:
                logging.info("enqueued reschedule failed, queue full")
                pass
        else:
            if not self._reschedule_queue.empty():
                self._reschedule_queue.get_nowait()
            await self._reschedule_queue.put(True)

    async def _reschedule_jobs(self):
        allocation = optimize(
            self._job, self._cluster, self._cluster_size)
        allocation = await self._cluster.expand_cluster(
            self._job.workers, allocation)

        worker_tasks = await self._job.update_workers(allocation)
        if worker_tasks:
            asyncio.create_task(
                self._handle_worker_failure(worker_tasks))
        self._cluster._force_immediate_allocation.clear()

    async def _spot_termination_handler(self, task):
        try:
            ip = await task
            self._cluster.mark_node_for_termination(ip)
            self._cluster._force_immediate_allocation.set()
            await self._enqueue_reschedule(immediate=True)
        except ray.exceptions.WorkerCrashedError:
            pass

    async def _handle_worker_failure(self, tasks):
        try:
            await asyncio.gather(*tasks)
        except Exception as e:
            if self._job._workers:
                logging.error("worker failure detected")
                logging.error(e)
                # Let the autoscheduler resolve any dead nodes
                await asyncio.sleep(60)
                # Todo: remove
                await self._enqueue_reschedule(immediate=True)

    async def register_worker(self, rank, ip):
        self._job._workers[rank] = ip
        if ip not in self._spot_listener_tasks:
            self._spot_listener_tasks[ip] = \
                listen_for_spot_termination.options(
                     num_cpus=0.1, resources={f"node:{ip}": 0.01}).remote()
            asyncio.create_task(
                self._spot_termination_handler(
                    self._spot_listener_tasks[ip]))

    async def register_checkpoint(self, checkpoint):
        self._job.register_checkpoint(checkpoint)
        return self._job._checkpoint_received.is_set()

    async def register_status(self, status):
        self._job.register_status(status)

    async def _handle_report(self, request):
        self._job.register_hints(await request.json())
        await self._enqueue_reschedule()
        return web.Response(text="metrics added")

    async def _handle_discover(self, request):
        # Long-polling endpoint used for discovering pod IPs for a given job.
        pod_ip_list = [
            ip for rank, ip in list(sorted(self._job.workers.items()))]
        # TODO: error handling
        return web.json_response(pod_ip_list)

    async def _run_app(self):
        port = 8080
        app = web.Application()
        app.add_routes([
            web.get('/discover/{namespace}/{name}/{group}',
                    self._handle_discover),
            web.put('/hints/{namespace}/{name}', self._handle_report),
        ])
        self._runner = web.AppRunner(app)
        await self._runner.setup()
        site = web.TCPSite(self._runner, services.get_node_ip_address(), port)
        await site.start()
        self._ready.set()
        return None


# For testing
_test_controller = Controller

# For general usage
Controller = ray.remote(num_cpus=1)(Controller)



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import ray

import json

import adaptdl_ray.aws.worker as worker
from aiohttp import web


@ray.remote
def mocked_listen_for_spot_termination():
    time.sleep(3)
    return "a different ip"


worker.listen_for_spot_termination = mocked_listen_for_spot_termination


from adaptdl_ray.aws.controller import RayAdaptDLJob, Cluster # noqa
from adaptdl_ray.aws.controller import _test_controller as Controller # noqa
from adaptdl_ray.aws.utils import Status # noqa

import time # noqa
import asyncio # noqa
from collections import namedtuple # noqa
import requests # noqa


# see https://stackoverflow.com/a/60334747
@pytest.fixture(scope="module")
def ray_fix():
    ray.init()
    yield None
    ray.shutdown()


MockedJob = namedtuple("Job", ["workers"])


async def test_adaptdl_job_checkpoint(ray_fix):
    job = RayAdaptDLJob(None, 0, 0)

    @ray.remote
    def worker():
        while True:
            time.sleep(1)
    tasks = {i: worker.remote() for i in range(5)}
    job._worker_tasks = tasks
    job._running = True
    await job.force_worker_checkpoint()
    assert not job._running


async def test_adaptdl_job_register_hints(ray_fix):
    job = RayAdaptDLJob(None, 0, 0)
    job.register_hints("some hints")
    assert job._last_metrics == "some hints"


async def test_adaptdl_job_hints(ray_fix):
    hints = {
        "gradParams": {"norm": 3.0, "var": 4.0},
        "perfParams": {
            'alpha_c': 0, 'beta_c': 0, 'alpha_n': 0,
            'beta_n': 0, 'alpha_r': 0, 'beta_r': 0, 'gamma': 0}}
    job = RayAdaptDLJob(None, 0, 0)
    job._last_metrics = hints
    job.hints


async def test_fetch_metrics():
    hints = {
        "gradParams": {"norm": 3.0, "var": 4.0},
        "perfParams": {
            'alpha_c': 0, 'beta_c': 0, 'alpha_n': 0,
            'beta_n': 0, 'alpha_r': 0, 'beta_r': 0, 'gamma': 0}}
    job = RayAdaptDLJob(None, 0, 0)
    job._last_metrics = hints
    job._fetch_metrics()


async def test_adaptdl_job_register_checkpoint(ray_fix):
    job = RayAdaptDLJob(None, 0, 0)
    checkpoint = "foo"
    assert not job._checkpoint_received.is_set()
    job.register_checkpoint(checkpoint)
    assert job._checkpoint == "foo"
    assert job._checkpoint_received.is_set()
    assert ray.get(job._checkpoint_ref) == "foo"


async def test_adaptdl_job_register_status(ray_fix):
    job = RayAdaptDLJob(None, 0, 0)
    status = Status.FAILED.value
    job.register_status(status)
    assert job._status == Status.FAILED
    assert job.completed.is_set()

    job = RayAdaptDLJob(None, 0, 0)
    status = Status.SUCCEEDED.value
    job.register_status(status)
    assert job._status == Status.SUCCEEDED
    assert job.completed.is_set()

    job = RayAdaptDLJob(None, 0, 0)
    status = Status.RUNNING.value
    job.register_status(status)
    assert job._status == Status.RUNNING
    assert not job.completed.is_set()


async def test_cluster_invalid_nodes(ray_fix):
    cluster = Cluster(None, 0)
    cluster.mark_node_for_termination("some ip")
    cluster.mark_node_for_termination("some other ip")
    assert (cluster._invalid_nodes ==
            {"some ip", "some other ip",
             ray._private.services.get_node_ip_address()})


async def test_controller_run(ray_fix):
    controller = Controller(100, 5)
    controller.app_ran = False

    class MockedRunner():
        def __init__(self):
            self.cleaned_up = False

        async def cleanup(self):
            self.cleaned_up = True

    async def mocked_run_app():
        controller._completed.set()
        controller._runner = MockedRunner()
        controller.app_ran = True

    controller._run_app = mocked_run_app
    await controller.run_controller()

    assert controller._runner.cleaned_up
    assert controller.app_ran
    await controller._runner.cleanup()


async def test_controller_create_job(ray_fix):
    controller = Controller(100, 5)
    controller._ready.set()
    controller.rescheduled = False

    async def mocked_reschedule():
        controller.rescheduled = True
        controller._job.completed.set()
        controller._job._status = Status.SUCCEEDED

    controller._reschedule_jobs = mocked_reschedule

    resources = {"CPU": 1, "GPU": 2}
    asyncio.create_task(controller._reschedule_listener())
    await controller.create_job(
        worker_resources=resources,
        worker_port_offset=0,
        checkpoint_timeout=1)

    assert controller._job._worker_resources == resources
    assert controller._job._worker_port_offset == 0
    assert controller._job._checkpoint_timeout == 1
    assert controller._cluster._worker_resources == resources
    assert controller.rescheduled
    assert controller._job._status == Status.SUCCEEDED


async def test_controller_reschedule_jobs(ray_fix):
    controller = Controller(100, 5)
    job = RayAdaptDLJob({"CPU": 1}, 0, 0)
    controller._job = job
    job.forced_checkpoint = False
    job.updated = 0
    controller.handled_workers = []

    async def mocked_handle_worker_failure(tasks):
        controller.handled_workers += tasks

    async def mocked_update_workers(allocation):
        await asyncio.sleep(3)
        job.updated += 1
        if job._workers != allocation:
            job._workers = allocation
            return allocation
        return None

    controller._handle_worker_failure = mocked_handle_worker_failure
    job.update_workers = mocked_update_workers

    controller._cluster = Cluster(None, 0)
    controller._cluster.expanded = None

    async def mocked_expand_cluster(workers, allocation):
        controller._cluster.expanded = allocation
        return allocation

    controller._cluster.expand_cluster = mocked_expand_cluster

    async def wrapped_call(duration):
        await asyncio.sleep(duration)
        await controller._reschedule_jobs()

    await asyncio.wait_for(
        asyncio.gather(
            wrapped_call(0), wrapped_call(1), wrapped_call(2)),
        15)

    await asyncio.sleep(4)
    assert job.updated == 3

    # Default allocation
    assert controller.handled_workers == ['adaptdl_virtual_node_0']
    assert controller._cluster.expanded == ['adaptdl_virtual_node_0']


async def test_controller_spot_termination_handler(ray_fix):
    controller = Controller(100, 5)
    job = RayAdaptDLJob(None, 0, 0)
    controller._job = job
    controller.rescheduled = False

    async def mocked_reschedule():
        controller.rescheduled = True

    controller._cluster = Cluster(None, 0)
    controller._reschedule_jobs = mocked_reschedule
    asyncio.create_task(controller._reschedule_listener())

    controller._cluster.marked = None

    def mocked_mark_node_for_termination(ip):
        controller._cluster.marked = ip

    controller._cluster.mark_node_for_termination = \
        mocked_mark_node_for_termination

    async def task():
        return "some ip"

    async def wrapper():
        awaitable_task = asyncio.create_task(task())
        await controller._spot_termination_handler(awaitable_task)

    await wrapper()
    await asyncio.sleep(4)
    assert controller.rescheduled
    assert controller._cluster.marked == "some ip"


async def test_controller_register_worker(ray_fix):
    controller = Controller(100, 5)
    job = RayAdaptDLJob(None, 0, 0)
    controller._job = job
    controller._spot_listener_tasks = {"some-ip": 1}

    controller.task_result = None

    async def mocked_spot_termination_handler(task):
        controller.task_result = ray.get(task)

    controller._spot_termination_handler = mocked_spot_termination_handler

    ip = ray._private.services.get_node_ip_address()

    await controller.register_worker(0, "some-ip")
    await controller.register_worker(
        1, ray._private.services.get_node_ip_address())

    await asyncio.sleep(1)

    assert job._workers[0] == "some-ip"
    assert job._workers[1] == ip
    assert controller.task_result == "a different ip"


async def test_controller_register_checkpoint(ray_fix):
    controller = Controller(100, 5)
    job = RayAdaptDLJob(None, 0, 0)
    controller._job = job
    checkpoint = "foo"
    checkpoint_received = await controller.register_checkpoint(checkpoint)
    assert checkpoint_received
    assert job._checkpoint_received
    assert job._checkpoint == "foo"
    assert ray.get(job._checkpoint_ref) == "foo"


async def test_controller_register_status():
    controller = Controller(100, 5)
    job = RayAdaptDLJob(None, 0, 0)
    controller._job = job
    status = Status.RUNNING.value
    await controller.register_status(status)
    assert(job._status == Status.RUNNING and not job.completed.is_set())
    status = Status.SUCCEEDED.value
    await controller.register_status(status)
    assert(job._status == Status.SUCCEEDED and job.completed.is_set())


async def test_controller_handle_report():
    controller = Controller(100, 1)
    job = RayAdaptDLJob(None, 0, 1)
    controller._job = job
    controller.rescheduled = False

    async def mocked_reschedule():
        controller.rescheduled = True

    controller._reschedule_jobs = mocked_reschedule
    asyncio.create_task(controller._reschedule_listener())

    class MockedRequest:
        def __init__(self, body):
            self._body = body

        async def json(self):
            return self._body
    hints = {"some": "hints"}
    hints_json = MockedRequest(json.dumps(hints))
    await controller._handle_report(hints_json)

    await asyncio.sleep(5)

    assert(
        controller.rescheduled and
        json.loads(job._last_metrics) == hints and
        id(job._last_metrics) != id(hints))


async def test_controller_handle_discover():
    controller = Controller(4, 100)
    controller._job = MockedJob(
        workers={0: "127.0.0.1", 1: "127.0.0.2", 2: "0.0.0.0"})
    workers = await controller._handle_discover(None)
    assert (json.loads(workers.text) == ["127.0.0.1", "127.0.0.2", "0.0.0.0"])


async def test_controller_run_app(aiohttp_client):
    controller = Controller(4, 100)
    controller.called_report = False
    controller.called_discover = False

    async def mocked_report(request):
        controller.called_report = True
        return web.Response(text="Success Hints")

    async def mocked_discover(request):
        controller.called_discover = True
        return web.Response(text="Success Discover")

    async def put():
        await asyncio.sleep(5)
        client = await aiohttp_client(controller._runner.app())
        return await client.put("/hints/namespace/name")

    async def get():
        await asyncio.sleep(5)
        client = await aiohttp_client(controller._runner.app())
        return await client.get("/discover/namespace/name/group")

    controller._handle_discover = mocked_discover
    controller._handle_report = mocked_report
    await controller._run_app()
    put_response = await put()
    get_response = await get()

    put_text = await(put_response.text())
    assert put_text == "Success Hints"
    get_text = await(get_response.text())
    assert get_text == "Success Discover"

    assert controller.called_report
    assert controller.called_discover
    await controller._runner.cleanup()



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Dict, List
from collections import defaultdict
import ray

_DEFAULT_DEVICE = None

# Job-level replica bounds
_JOB_MIN_REPLICAS = 0
_JOB_MAX_REPLICAS = 10


def _avail_nodes() -> Dict:
    """ We return all live nodes with their allocatable resources."""
    alive_nodes = {node["NodeID"]: node for node in
                   sorted(ray.nodes(), key=lambda x: x['NodeID'])
                   if node["alive"]}
    for node_id, resources in ray.state.state.\
            _available_resources_per_node().items():
        pruned_resources = defaultdict(float, {k: v for k, v in
                                       resources.items() if "group" not in k})
        alive_nodes[node_id]["Resources"] = pruned_resources
    return alive_nodes


def default_device(refresh=False) -> str:
    """ Default device will be GPU if at least one node has a GPU on it else we
    use CPUs. It is initialized once when the allocator/scheduler is
    instantiated."""
    global _DEFAULT_DEVICE
    if _DEFAULT_DEVICE is None or refresh:
        assert ray.is_initialized()
        if any("GPU" in node['Resources'] for node in _avail_nodes().values()):
            _DEFAULT_DEVICE = "GPU"
        else:
            _DEFAULT_DEVICE = "CPU"
    return _DEFAULT_DEVICE


def job_resources():
    """ Default job resources."""
    return {default_device(): 1, "memory": 1024*1024}


def nodes(consumed_resources: Dict = None) -> List[Dict]:
    """Returns all nodes with the default device on them"""
    avail_nodes = _avail_nodes()
    if consumed_resources is not None:
        # Add resources we are consuming already to the available resources
        node_id_map = {node["NodeID"]: node["NodeManagerAddress"] for node in
                       ray.nodes() if node["alive"]}
        for node_id, node in avail_nodes.items():
            resources = consumed_resources.get(node_id_map[node_id], {})
            for resource, amount in resources.items():
                node["Resources"][resource] += amount
    return [node for node in avail_nodes.values()
            if default_device() in node["Resources"]]



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List
from collections import Counter, defaultdict
from copy import deepcopy
from ray import tune
from ray.util.placement_group import get_current_placement_group
from adaptdl_ray.adaptdl import config


def pgf_to_allocation(pgf) -> List[str]:
    """ Convert a Placement Groups Factory to AdaptDL allocation"""
    bundles = pgf._bundles[1:]
    allocs, node_keys, num_devices = [], [], []
    for bundle in bundles:
        node_keys += [k.split(":")[1] for k, v in bundle.items()
                      if k.startswith("node")]
        num_devices += [int(v) for k, v in bundle.items()
                        if k == config.default_device()]

    for node, count in zip(node_keys, num_devices):
        allocs += [node] * count
    return allocs


def allocation_to_pgf(alloc: List[str], resources_per_node=None):
    """ Convert AdaptDL allocation to a Placement Group Factory"""
    if not resources_per_node:
        resources_per_node = {"CPU": 1.0}
        if config.default_device() == "GPU":
            resources_per_node["GPU"] = 1.0

    def _construct_bundle(node, number_of_instances):
        resources = deepcopy(resources_per_node)
        resources["CPU"] *= number_of_instances
        if "GPU" in resources:
            resources["GPU"] *= number_of_instances
        if "adaptdl_virtual" not in node:
            resources[f"node:{node}"] = 0.01
        return resources

    assert len(alloc) > 0
    resources = [{"CPU": 0.001}]
    alloc = Counter(alloc)
    for node, res in alloc.items():
        resources.append(_construct_bundle(node, res))
    return tune.PlacementGroupFactory(resources)


def pgf_to_num_replicas(pgf) -> int:
    """ Extract the number of replicas of the trial from its PGF"""
    return sum(int(bundle.get(config.default_device(), 0))
               for bundle in pgf._bundles[1:])


def pgs_to_resources(pgs: List[Dict]) -> Dict:
    """ Return node-level resource usage by all PGs in pgs."""
    # Note that every bundle is tagged with the node resource
    resources = defaultdict(Counter)
    for pg in pgs:
        for bundle in pg["bundle_cache"][1:]:
            # Every bundle has a node resource
            node_ip = [k.split(":")[1] for k in bundle.keys()
                       if k.startswith("node")][0]
            for k, v in bundle.items():
                resources[node_ip][k] += v
    return resources


def unique_nodes_pg() -> int:
    nodes = []
    if get_current_placement_group() is None:
        return 0
    else:
        for bundle in get_current_placement_group().bundle_specs:
            for resource in bundle:
                if "node" in resource:
                    nodes.append(resource)
        return len(set(nodes))



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import logging
import os

from adaptdl_ray.aws.controller import Controller
from adaptdl_ray.aws.utils import Status

import ray


logging.basicConfig()
LOG = logging.getLogger(__name__)
LOG.setLevel(logging.INFO)


def run_adaptdl_on_ray_cluster(
        path, argv, ray_uri, working_dir,
        worker_resources, cluster_size, worker_port_offset,
        checkpoint_timeout, rescale_timeout):
    LOG.info("Starting AdaptDLJob")
    if ray.is_initialized():
        return
    if not os.path.exists(working_dir):
        raise RuntimeError(f"Cannot find local directory {working_dir}")
    if not os.path.exists(os.path.join(working_dir, path)):
        raise RuntimeError(
            f"Cannot find local file {os.path.join(working_dir, path)}")
    runtime_env = {
        "working_dir": working_dir}
    ray.init(ray_uri, runtime_env=runtime_env)

    controller = Controller.options(name="AdaptDLController").remote(
        cluster_size, rescale_timeout)

    controller.run_controller.remote()
    try:
        status_obj = controller.create_job.remote(
            worker_resources, worker_port_offset, checkpoint_timeout,
            path=path, argv=argv)
        status = ray.get(status_obj)
        if status.value == Status.SUCCEEDED.value:
            LOG.info("Job succeeded")
            return 0
        else:
            raise RuntimeError("Job failed")
    except Exception as e:
        raise e


def main():
    parser = argparse.ArgumentParser(description="Adaptdl on Ray")
    parser.add_argument(
        "-f", "--file", type=str,
        help=("File to run on the cluster. The path must be a relative path"
              "rooted at the argument of --working-dir"), required=True)
    parser.add_argument(
        "-u", "--uri", type=str,
        help="URI of the ray cluster, e.g. `ray://<ip>:10001", required=True)
    parser.add_argument(
        "-m", "--max-cluster-size", type=int,
        help="Maximum number of workers (with sufficient gpus) in the cluster",
        required=True)
    parser.add_argument(
        "--gpus", type=int, help="number of gpus per worker", default=1)
    parser.add_argument(
        "--cpus", type=int, help="number of cpus per worker", default=1)
    parser.add_argument(
        "--port-offset", type=int, default=0,
        help=("Torch communication worker port offset. Generally to be used "
              "in case the adaptdl communication ports are taken "))
    parser.add_argument(
        "-d", "--working-dir", type=str,
        help=("Directory to copy to the worker tasks. Should contain the file "
              "specified by -f/--file."),
        required=True)
    parser.add_argument(
        "--checkpoint-timeout", type=int, default=120,
        help=("Number of seconds that the controller will wait for the "
              "checkpoint object to be reported back by worker 0. "
              "If the checkpoint is not received by that time, an old "
              "checkpoint will be used, if one exists"))
    parser.add_argument(
        "--cluster-rescale-timeout", type=int, default=60,
        help=("Number of seconds that the controller will wait for the "
              "cluster to rescale to the desired size. If the desired size "
              "is not reached by the end of this timeout, then the controller "
              "will allocate on whatever nodes it finds then"))
    parser.add_argument(
        "arguments", type=str, nargs=argparse.REMAINDER,
        help=("Command line arguments to be passed to the file specified by "
              "-f/--f. You many seperate these from the other command line "
              "arguments with --"))
    args = parser.parse_args()
    if args.arguments and args.arguments[0] == "--":
        arguments = args.arguments[1:]
    else:
        arguments = []
    run_adaptdl_on_ray_cluster(
        args.file,
        argv=arguments,
        ray_uri=args.uri,
        working_dir=args.working_dir,
        worker_resources={"CPU": args.cpus, "GPU": args.gpus},
        cluster_size=args.max_cluster_size,
        worker_port_offset=args.port_offset,
        checkpoint_timeout=args.checkpoint_timeout,
        rescale_timeout=args.cluster_rescale_timeout)


if __name__ == "__main__":
    main()



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Callable, Dict, Optional
from unittest.mock import patch
from datetime import timedelta

import ray
from ray.tune.resources import Resources
from ray.tune.registry import register_trainable
from ray.tune.integration.torch import _TorchTrainable
from ray.util.sgd.torch.constants import NCCL_TIMEOUT_S

import adaptdl_ray.tune.adaptdl_patch as P
from adaptdl_ray.adaptdl import config


def AdaptDLTrainableCreator(func: Callable,
                            num_workers: int = 1,
                            group: int = 0,
                            num_cpus_per_worker: int = 1,
                            num_workers_per_host: Optional[int] = None,
                            backend: str = "gloo",
                            timeout_s: int = NCCL_TIMEOUT_S,
                            use_gpu=None):
    """ Trainable creator for AdaptDL's elastic Trials"""
    if config.default_device() == "GPU":
        backend = "nccl"

    class AdaptDLTrainable(_TorchTrainable):
        """ Similar to DistributedTrainable but for AdaptDLTrials."""
        def setup(self, config: Dict):
            """ Delay-patch methods when the Trainable actors are first
            created"""
            with patch(target="ray.tune.integration.torch.setup_process_group",
                       new=P.setup_process_group), \
                 patch(target='ray.tune.integration.torch.wrap_function',
                       new=P.wrap_function_patched):
                _TorchTrainable.setup(self, config)

        # Override the default resources and use custom PG factory
        @classmethod
        def default_resource_request(cls, config: Dict) -> Resources:
            return None

        def get_sched_hints(self):
            return ray.get(self.workers[0].get_sched_hints.remote())

        def save_all_states(self, trial_state):
            return ray.get(self.workers[0].save_all_states.remote(trial_state))

        @classmethod
        def default_process_group_parameters(self) -> Dict:
            return dict(timeout=timedelta(timeout_s), backend=backend)

    AdaptDLTrainable._function = func
    AdaptDLTrainable._num_workers = num_workers
    # Set number of GPUs if we're using them, this is later used when spawning
    # the trial actors
    if config.default_device() == "GPU":
        AdaptDLTrainable._num_gpus_per_worker = 1
    else:
        AdaptDLTrainable._num_gpus_per_worker = 0

    # Trainables are named after number of replicas they spawn. This is
    # essential to associate the right Trainable with the right Trial and PG.
    AdaptDLTrainable.__name__ = AdaptDLTrainable.__name__.split("_")[0] + \
        f"_{num_workers}" + f"_{group}"
    register_trainable(AdaptDLTrainable.__name__, AdaptDLTrainable)
    return AdaptDLTrainable


# To support unit tests and integration tests
def _train_simple(config: Dict, checkpoint_dir: Optional[str] = None):
    import torch
    import torch.nn as nn
    import torch.optim as optim
    import adaptdl.torch as adl
    from ray import tune

    class MyDataset:
        def __init__(self, xs, ys):
            self.xs = xs
            self.ys = ys

        def __getitem__(self, i):
            return self.xs[i], self.ys[i]

        def __len__(self):
            return len(self.xs)

    # N is batch size; D_in is input dimension;
    # H is hidden dimension; D_out is output dimension.
    N, D_in, H, D_out = 64, 5, 5, 5
    dataset = MyDataset(torch.randn(N, D_in), torch.randn(N, D_out))

    H = config.get("H", 16)
    N = config.get("N", 16)

    # Create random Tensors to hold inputs and outputs
    dataloader = adl.AdaptiveDataLoader(dataset, batch_size=N)
    dataloader.autoscale_batch_size(4096, local_bsz_bounds=(16, 1024))

    loss_fn = nn.MSELoss()

    # Use the nn package to define our model and loss function.
    model = torch.nn.Sequential(
        torch.nn.Linear(D_in, H),
        torch.nn.ReLU(),
        torch.nn.Linear(H, D_out),
    )
    optimizer = optim.SGD(model.parameters(), lr=0.1)

    model = adl.AdaptiveDataParallel(model, optimizer)

    loss = torch.Tensor([0.0])
    for epoch in adl.remaining_epochs_until(config.get("epochs", 10)):
        for (x, y) in dataloader:
            optimizer.zero_grad()
            output = model(x)
            loss = loss_fn(output, y)
            loss.backward()
            optimizer.step()

        tune.report(mean_loss=loss.item())



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime
import logging
from typing import List

import ray
from ray.tune import trial_runner
from ray.tune.trial import Trial
from ray.tune import PlacementGroupFactory
from ray.tune.function_runner import FuncCheckpointUtil
from ray.tune.trainable import TrainableUtil

from adaptdl_ray.adaptdl import AdaptDLJobMixin
from adaptdl_ray.tune.adaptdl_trainable import AdaptDLTrainableCreator
from adaptdl_ray.adaptdl.utils import pgf_to_num_replicas, allocation_to_pgf

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class AdaptDLTrial(AdaptDLJobMixin, Trial):
    """ Tune Trial that brings in AdaptDL functionality. """
    def __init__(self, *args, **kwargs):
        self.rescale_count = kwargs.pop("rescale_count", 0)
        self._cached_metrics = None
        super().__init__(job_id=kwargs["trial_id"], *args, **kwargs)

    @property
    def _num_replicas(self) -> int:
        return self.get_trainable_cls()._num_workers

    def __getstate__(self):
        copy_state = {}
        # Remove problematic members
        for k in ("_trial_in_use", "_cached_metrics"):
            copy_state[k] = self.__dict__[k]
            del self.__dict__[k]
        state = super().__getstate__()
        # Restore members
        for k, v in copy_state.items():
            self.__dict__[k] = v
        return state

    def _requeue(self,
                 old_trial: Trial,
                 trial_runner: "trial_runner.TrialRunner"):
        # Remove the old trial from trial_runner
        trial_runner.trial_executor.stop_trial(old_trial)
        trial_runner._trials.pop(trial_runner._trials.index(old_trial))
        # Important: Add the new trial to the runner
        trial_runner._trials.append(self)
        trial_runner._live_trials.add(self)

    def _fetch_metrics(self):
        if self.runner is not None:
            metrics = ray.get(self.runner.get_sched_hints.remote())
            if metrics is not None:
                # Update cache
                self._cached_metrics = metrics
        return self._cached_metrics

    def _allocation_in_use(self):
        return self._trial_in_use(self)

    @classmethod
    def _clone_from(cls,
                    trial: Trial,
                    allocation,
                    restore_path=None) -> "AdaptDLTrial":
        trainable_cls = trial.get_trainable_cls()
        pgf = allocation_to_pgf(allocation)
        num_workers = pgf_to_num_replicas(pgf)
        assert num_workers > 0
        if isinstance(trial, AdaptDLTrial):
            # Cloning from existing AdaptDLTrial
            rescale_count = trial.rescale_count + 1
            # Carry over the creation_timestamp
            creation_timestamp = trial.creation_timestamp
        else:
            creation_timestamp = datetime.now()
            rescale_count = 0

        adaptdl_trainable_cls = AdaptDLTrainableCreator(trainable_cls.
                                                        _function,
                                                        num_workers,
                                                        group=rescale_count)
        return cls(trainable_name=adaptdl_trainable_cls.__name__,
                   creation_timestamp=creation_timestamp,
                   rescale_count=rescale_count,
                   config=trial.config,
                   experiment_tag=trial.experiment_tag,
                   evaluated_params=trial.evaluated_params,
                   stopping_criterion=trial.stopping_criterion,
                   trial_id=trial.trial_id,
                   restore_path=restore_path,
                   local_dir="/tmp",  # TODO: Decide a proper way
                   placement_group_factory=pgf)

    @classmethod
    def create_from(cls,
                    trial: Trial,
                    trial_runner: "trial_runner.TrialRunner",
                    new_allocation: List[str],
                    copy_state=False) -> "AdaptDLTrial":
        """ Create a new AdaptDLTrial from a Trial or AdaptDLTrial with new
        allocations. This also replaces the existing Trial."""
        checkpoint_path = None
        logger.debug(f"Creating {trial} with {len(new_allocation)} replicas")
        if copy_state:
            if trial.runner is not None:
                # Fetch the state from the other trial
                checkpoint_obj = ray.get(trial.runner.save_all_states.remote(
                                         trial.runner.get_state.remote()))
                # Dump it to disk
                temp_checkpoint_dir = (FuncCheckpointUtil.
                                       mk_temp_checkpoint_dir(trial.logdir))
                checkpoint_path = TrainableUtil. \
                    create_from_pickle(checkpoint_obj, temp_checkpoint_dir)
            else:
                # trial was PAUSED
                checkpoint_path = trial.restore_path

        # Spawn a new trial
        new_trial = cls._clone_from(trial, new_allocation,
                                    restore_path=checkpoint_path)
        # Keep it for later use by the trials
        new_trial._trial_in_use = trial_runner.trial_executor.\
            _pg_manager.trial_in_use
        # Replace with old trial
        new_trial._requeue(trial, trial_runner)
        assert new_trial.restore_path == checkpoint_path
        assert new_trial.status == Trial.PENDING
        return new_trial

    def pause(self, trial_runner):
        """ Pause the AdaptDLTrial with a checkpoint. We try to remove the PG
        attached to this trial"""
        assert self.runner is not None
        checkpoint_obj = ray.get(self.runner.save_all_states.remote(
                                 self.runner.get_state.remote()))
        # Serialize to disk
        temp_checkpoint_dir = (FuncCheckpointUtil.
                               mk_temp_checkpoint_dir(self.logdir))
        checkpoint_path = TrainableUtil.create_from_pickle(checkpoint_obj,
                                                           temp_checkpoint_dir)

        # Trial will be restored from the checkpoint_path when it's resumed
        self.restore_path = checkpoint_path

        # Clear the allocation. This is a hack to clear the PG associated with
        # the trial. We assign a temporary PG which will get replaced with a
        # real PG once we resume the trial. This is needed because Tune likes
        # to keep the PGs around even for PAUSED trials.
        self.placement_group_factory = PlacementGroupFactory([{"CPU": 0.001}])
        # This forces Tune to garbage-collect uneeded PGs which can then be
        # reused
        trial_runner.trial_executor._pg_manager.\
            reconcile_placement_groups([self])
        logger.debug(f"PAUSING {self} w/ checkpoint at {checkpoint_path}")



# Copyright 2021 Petuum, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import adaptdl_ray.aws.worker as worker
import ray.autoscaler.sdk as sdk
import ray


class MockedRunAdaptDL:
    @staticmethod
    def options(*args, **kwargs):
        class InnerClass:
            def __init__(self):
                self.called = False
                self.args = None
                self.kwargs = None

            def remote(self, *args, **kwargs):
                self.called = True
                self.args = args
                self.kwargs = kwargs
                return self
        return InnerClass()


worker.run_adaptdl = MockedRunAdaptDL


def mocked_nodes():
    current_ip = ray._private.services.get_node_ip_address()
    return [
        {"NodeManagerAddress": current_ip,
         "alive": True,
         "Resources": "some value"},
        {"NodeManagerAddress": "some ip address",
         "alive": True},
        {"NodeManagerAddress": "some other ip address",
         "alive": False,
         "Resources": "some value"},
        {"NodeManagerAddress": "one last ip address",
         "alive": True,
         "Resources": {"CPU": 2}}]


worker.bundles = None


def mocked_sdk_request(bundles=None, *args, **kwargs):
    worker.bundles = bundles


sdk.request_resources = mocked_sdk_request
ray.nodes = mocked_nodes


import adaptdl_ray.aws.controller as controller# noqa

controller.FULL_RESCALE_TIMEOUT = 4

from adaptdl_ray.aws.controller import Cluster, RayAdaptDLJob # noqa


async def test_adaptdl_job_create_workers():
    job = RayAdaptDLJob({"CPU": 2}, 0, 0)

    async def mocked_handle_worker_failure(tasks):
        pass

    job._handle_worker_failure = mocked_handle_worker_failure
    await job._create_workers([
        "adaptdl_virtual_node_0", "adaptdl_virtual_node_1",
        "adaptdl_virtual_node_2"])
    assert len(job._worker_tasks) == 3
    for i in range(3):
        print(type(job._worker_tasks[i]))
        assert job._worker_tasks[i].called


async def test_adaptdl_job_update_workers():
    job = RayAdaptDLJob({"CPU": 2}, 0, 3)
    allocation = [
        "adaptdl_virtual_node_0", "adaptdl_virtual_node_1",
        "adaptdl_virtual_node_2"]

    async def mocked_force_worker_checkpoint():
        print("forcing checkpoint")
        job._checkpoint_received = True
        job._checkpoint = 3

    async def mocked_create_workers(allocation):
        job._workers_created = True
        job._worker_ips = allocation

    job.force_worker_checkpoint = mocked_force_worker_checkpoint
    job._workers_created = False
    job._worker_ips = None
    job._create_workers = mocked_create_workers

    await job.update_workers(allocation)
    assert job._checkpoint == 3
    assert job._workers_created
    assert job._worker_ips == allocation

    job._checkpoint = 0
    job._workers_created = False
    job._workers = {i: value for i, value in enumerate(allocation)}
    await job.update_workers(allocation)
    assert not job._workers_created
    assert job._checkpoint == 0


async def test_cluster_get_nodes():
    cluster = Cluster(None, 0)
    cluster.mark_node_for_termination("some ip address")
    nodes = cluster.get_nodes()
    assert len(nodes) == 1
    assert nodes[0]["NodeManagerAddress"] == "one last ip address"


async def test_cluster_ready():
    cluster = Cluster({"CPU": 1}, 5)
    allocation_1 = ["one last ip address"]
    allocation_2 = ["one last ip address", "an ip we don't have"]
    allocation_3 = ["one last ip address", "an ip we don't have",
                    "one ip too many"]

    found, count = cluster._cluster_ready(allocation_1)
    assert found
    assert count == 1

    found, count = cluster._cluster_ready(allocation_2)
    assert found
    assert count == 2

    found, count = cluster._cluster_ready(allocation_3)
    assert not found
    assert count == 2


async def test_expand_cluster():
    cluster = Cluster({"CPU": 1}, 3)
    # Note: placeholders will have "adaptdl_virtual" in their name
    allocation = [
        "an adaptdl_virtual ip we don't have",
        "one last real ip address",
        "one adaptdl_virtual ip too many"]
    result = await cluster.expand_cluster({}, allocation)
    assert len(worker.bundles) == 3
    assert "one last real ip address" in result
    assert len(result) == 2

    allocation = [
        "an adaptdl_virtual ip we don't have",
        "one last real ip address"]
    result = await cluster.expand_cluster({}, allocation)
    assert len(worker.bundles) == 2
    assert "one last real ip address" in result
    assert len(result) == 2
